print("import: 'emirdrp'")
import emirdrp

print("import: 'emirdrp.core'")
import emirdrp.core

print("import: 'emirdrp.instrument'")
import emirdrp.instrument

print("import: 'emirdrp.instrument.configs'")
import emirdrp.instrument.configs

print("import: 'emirdrp.instrument.tests'")
import emirdrp.instrument.tests

print("import: 'emirdrp.processing'")
import emirdrp.processing

print("import: 'emirdrp.processing.tests'")
import emirdrp.processing.tests

print("import: 'emirdrp.processing.wavecal'")
import emirdrp.processing.wavecal

print("import: 'emirdrp.recipes'")
import emirdrp.recipes

print("import: 'emirdrp.recipes.acquisition'")
import emirdrp.recipes.acquisition

print("import: 'emirdrp.recipes.aiv'")
import emirdrp.recipes.aiv

print("import: 'emirdrp.recipes.aiv.tests'")
import emirdrp.recipes.aiv.tests

print("import: 'emirdrp.recipes.auxiliary'")
import emirdrp.recipes.auxiliary

print("import: 'emirdrp.recipes.calibration'")
import emirdrp.recipes.calibration

print("import: 'emirdrp.recipes.engineering'")
import emirdrp.recipes.engineering

print("import: 'emirdrp.recipes.image'")
import emirdrp.recipes.image

print("import: 'emirdrp.recipes.image.tests'")
import emirdrp.recipes.image.tests

print("import: 'emirdrp.recipes.spec'")
import emirdrp.recipes.spec

print("import: 'emirdrp.recipes.spec.tests'")
import emirdrp.recipes.spec.tests

print("import: 'emirdrp.recipes.tests'")
import emirdrp.recipes.tests

print("import: 'emirdrp.simulation'")
import emirdrp.simulation

print("import: 'emirdrp.tests'")
import emirdrp.tests

print("import: 'emirdrp.tools'")
import emirdrp.tools

print("import: 'emirdrp.util'")
import emirdrp.util

